#include <gtk/gtk.h>
#include <gdk/gdkx.h>
#include <libgnome/gnome-desktop-item.h>

#include "app-shell-startup.h"
#include "app-shell.h"
#include "search-bar.h"
#include "slab-section.h"

gint apss_new_instance_cb (BonoboApplication *app, gint argc, char *argv[], gpointer data)
{
    //g_message ("new_instance_cb got called\n");
    AppShellData * app_data = (AppShellData *)data;

    SlabSection * section = SLAB_SECTION(app_data->filter_section);
    g_assert(NLD_IS_SEARCH_BAR(section->contents) );  //Make sure our implementation has not changed
    NldSearchBar * search_bar = NLD_SEARCH_BAR(section->contents);

    gboolean visible;
    g_object_get(app_data->main_gnome_app, "visible", &visible, NULL);
    if(!visible)
    {
        nld_search_bar_set_text(search_bar, "", TRUE);
        show_shell(app_data);
    }

    if(argc) //if we are passed a valid startup time
    {
        gchar ** results = g_strsplit(argv[0], "_TIME", 0);
        gint lastentry = 0;
        while(results[lastentry] != NULL)
            lastentry++;
        guint32 timestamp = (guint32) g_strtod(results[lastentry-1], NULL);
        //printf("converted time stamp is %d\n", timestamp);
        g_strfreev(results);

        //gdk_x11_window_move_to_current_desktop(window);
        gdk_x11_window_set_user_time(app_data->main_gnome_app->window, timestamp);
    }

    //gtk_window_present_with_time(GTK_WINDOW(app_data->main_gnome_app), gtk_get_current_event_time());
    gtk_window_present(GTK_WINDOW(app_data->main_gnome_app));
    gtk_widget_grab_focus(GTK_WIDGET(search_bar));

    //gdk_notify_startup_complete();
	return argc;
}

gboolean apss_already_running(int argc, char *argv [], BonoboApplication ** app, const gchar * name)
{
    //if(bonobo_is_initialized())
    if(bonobo_init(&argc, argv) == FALSE)
        g_error ("Problem with bonobo_init");
    if(!bonobo_activate ())
        g_error("Problem with bonobo_activate()");

    bonobo_activation_set_activation_env_value("DISPLAY",
        gdk_display_get_name(gdk_display_get_default()));

    *app = bonobo_application_new(name);

    gchar const *envp[] = { "LANG", NULL };
    gchar * serverinfo = bonobo_application_create_serverinfo (*app, envp);

    BonoboAppClient *client;

    Bonobo_RegistrationResult reg_res = bonobo_application_register_unique(*app, serverinfo, &client);
    g_free (serverinfo);

    switch (reg_res)
    {
    case Bonobo_ACTIVATION_REG_ALREADY_ACTIVE: {
        int                           i;

        //g_message ("I am an application client.");
        bonobo_object_unref (BONOBO_OBJECT (*app));
        *app = NULL;

        gchar * newargv[1];
        newargv[0] = g_strdup(g_getenv("DESKTOP_STARTUP_ID"));
        i = bonobo_app_client_new_instance (client, ((newargv[0] && newargv[0][0] != '\0') ? 1 : 0), newargv, NULL);
        //g_message ("new-instance returned %i", i);
        g_object_unref (client);
        //gdk_notify_startup_complete();
        //bonobo_debug_shutdown ();
        return TRUE;

    }
    case Bonobo_ACTIVATION_REG_SUCCESS:
        //g_message ("I am an application server");
        //g_signal_connect (*app, "message::test-message", G_CALLBACK (message_cb), NULL);
        //g_signal_connect (*app, "new-instance", G_CALLBACK (new_instance_cb), app_data);
        //bonobo_application_new_instance (*app, argc, argv);
        //g_signal_connect (*app, "message::quit", G_CALLBACK (message_quit_cb), NULL);
        return FALSE;
        break;

    case Bonobo_ACTIVATION_REG_ERROR:
    default:
        g_error("bonobo activation error when registering unique application");
        return FALSE;
    }
}

